"""Define a bunch of functions that change the request.

These functions determine the handelability of a request, and run in the order
given here.

"""
import logging
import os
import urllib
import urlparse
from os.path import join, isfile, isdir, dirname, exists

from aspen import Response


log = logging.getLogger('aspen.gauntlet')


def intercept_socket(request):
    """Given a request object, return a tuple of (str, None) or (str, str).

    Intercept socket requests. We modify the filesystem path so that your
    application thinks the request was to /foo.sock instead of to
    /foo.sock/blah/blah/blah/.

    """
    if request.path.raw.endswith('.sock'):
        # request.path.raw does not include querystring.
        raise Response(404)
    parts = request.path.raw.rsplit('.sock/', 1)
    if len(parts) == 1:
        path = parts[0]
        socket = None
    else:
        path = parts[0] + '.sock'
        socket = parts[1]
    request.path.raw, request.socket = path, socket
    #spam -- log.debug('gauntlet.intercept_socket: ' + request.path.raw)

def translate(request):
    """Translate urlpath to fspath, returning urlpath parts.

    We specifically avoid removing symlinks in the path so that the filepath
    remains under the website root. Also, we don't want trailing slashes for
    directories in request.fs.

    """
    parts = [request.root] + request.path.raw.lstrip('/').split('/')
    request.fs = os.sep.join(parts).rstrip(os.sep)
    request._parts = parts # store for use in processing virtual_paths
    #spam -- log.debug('gauntlet.translate: ' + request.fs)

def check_sanity(request):
    """Make sure the request is under our root.
    """
    if not request.fs.startswith(request.root):
        raise response(404)

def hidden_files(request):
    """Protect hidden files.
    """
    if '/.' in request.fs[len(request.root):]:
        raise Response(404)

def virtual_paths(request):
    """Support /foo/bar.html => ./%blah/bar.html and /blah.html => ./%flah.html

    Parts is a list of fs path parts as returned by translate, above. 

    Path parts will end up in request.path, a dict subclass. There can only be 
    one variable per path part. If a directory has more than one subdirectory
    starting with '%' then only the 'first' is used.

    """
    if os.sep + '%' in request.fs[len(request.root):]:  # disallow direct access
        raise Response(404)
    if not exists(request.fs):
        matched = request.root
        parts = request._parts
        del request._parts
        nparts = len(parts)
        for i in range(1, nparts):
            part = parts[i]
            next = join(matched, part)
            if exists(next):    # this URL part names an actual directory
                matched = next
            else:               # this part is missing; do we have a %subdir?
                key = None
                for root, dirs, files in os.walk(matched):
                    files.sort(key=lambda x: x.lower())
                    dirs.sort(key=lambda x: x.lower())
                    for name in files + dirs:
                        if name.startswith('%'):
                            
                            # See if we can use this item.
                            # ============================
                            # We want to allow file matches for the last URL
                            # path part, and in that case we strip the file
                            # extension. For other matches we need them to be
                            # directories.

                            fs = join(matched, name)
                            k = name[1:]
                            v = part
                            if i == (nparts - 1):
                                if isfile(fs):
                                    # Take care with extensions.
                                    x = k.rsplit('.', 1)
                                    y = part.rsplit('.', 1)
                                    nx = len(x) # 1|2
                                    if nx != len(y):
                                        continue
                                    if nx == 2:
                                        # If there's an extension, match it.
                                        k, ext1 = x
                                        v, ext2 = y
                                        if ext1 != ext2:
                                            continue
                            elif not isdir(fs):
                                continue 


                            # We found a suitable match at the current level.
                            # ===============================================

                            matched = fs 
                            key, value = _typecast(k, v)
                            request.path[key] = value
                            break # Only use the first %match per level.
                    break # don't recurse in os.walk
                if key is None:
                    matched = request.root
                    break # no match, reset
        if matched != request.root:
            request.fs = matched.rstrip(os.sep)
    #spam -- log.debug('gauntlet.virtual_paths: ' + request.fs)

def _typecast(key, value):
    """Given two strings, return a string, and an int or string.
    """
    if key.endswith('.int'):    # you can typecast to int
        key = key[:-4]
        try:
            value = int(value)
        except ValueError:
            raise Response(404)
    else:                       # otherwise it's URL-quoted ASCII
        try:
            value = value.decode('ASCII')
        except UnicodeDecodeError:
            raise Response(400)
        value = urllib.unquote(value)
    return key, value

def trailing_slash(request):
    if isdir(request.fs):
        if not request.path.raw.endswith('/'):
            request.path.raw += '/'
            raise Response(301, headers={'Location': request.rebuild_url()})

def index(request):
    if isdir(request.fs):
        for filename in request.default_filenames:
            index = join(request.fs, filename)
            if isfile(index):
                request.fs = index
                break
    #spam -- log.debug('gauntlet.index: ' + request.fs)

def autoindex(request):
    if isdir(request.fs):
        if request.conf.aspen.no('list_directories'):
            request.headers.set('X-Aspen-AutoIndexDir', request.fs)
            request.fs = request.website.ours_or_theirs('autoindex.html')
            assert request.fs is not None # sanity check
        else:
            raise Response(404)
    #spam -- log.debug('gauntlet.autoindex: ' + request.fs)

def not_found(request):
    if not isfile(request.fs):
        if request.path.raw == '/favicon.ico': # special case
            request.fs = request.website.find_ours('favicon.ico')
        else:
            raise Response(404)
    #spam -- log.debug('gauntlet.not_found: ' + request.fs)


gauntlet = [ intercept_socket
           , translate
           , check_sanity
           , hidden_files
           , virtual_paths
           , trailing_slash
           , index
           , autoindex
           , not_found
            ]

def run(request):
    """Given a request, run it through the gauntlet.
    """
    #spam -- log.debug('gauntlet.run: ' + request.path.raw)
    for func in gauntlet:
        func(request)

def run_through(request, last):
    """For testing, here's a function that runs part of the gauntlet.

    Pass in a request object and a gauntlet function, the last to be run.

    """
    #spam -- log.debug('gauntlet.run_through: ' + request.path.raw)
    for func in gauntlet:
        func(request)
        if func is last:
            break



class Resource(object):
    """This is a base class for both static and dynamic resources.
    """

    def __init__(self, website, fs, raw, mimetype, modtime):
        self.website = website
        self.fs = fs
        self.raw = raw
        self.mimetype = mimetype
        self.modtime = modtime



"""Aspen uses Resource classes to model HTTP resources.

Here is the class hierarchy:

    Resource ------> DynamicResource ------> JSONResource
             \                       \ 
              \                       \----> SocketResource
               \                       \
                \--> StaticResource     \--> TemplateResource

XXX: There is lots of semantic ambiguity and overlap between Resource and
Simplate.

"""
import datetime
import logging
import mimetypes
import os
import stat
import sys
import traceback
from os.path import join

PAGE_BREAK = chr(12) # used in the following imports

from aspen import json
from aspen.exceptions import LoadError
from aspen.resources.json_resource import JSONResource
from aspen.resources.template_resource import TemplateResource
from aspen.resources.socket_resource import SocketResource
from aspen.resources.static_resource import StaticResource


log = logging.getLogger('aspen.resources')


# Cache helpers
# =============

__cache__ = dict()        # cache, keyed to filesystem path

class Entry:
    """An entry in the global resource cache.
    """

    fspath = ''         # The filesystem path [string]
    modtime = None      # The timestamp of the last change [int]
    quadruple = None    # A post-processed version of the data [4-tuple]
    exc = None          # Any exception in reading or compilation [Exception]

    def __init__(self):
        self.fspath = ''
        self.modtime = 0
        self.quadruple = ()


# Core loaders
# ============

def get_resource_class(raw, mimetype):
    """Given raw file contents and a mimetype, return a Resource subclass.

    This function encodes the algorithm for deciding what kind of Resource (==
    simplate) a given file is. Is it a static file or a dynamic JSON simplate
    or what? Etc. Here is the algorithm:

        If mimetype is 'application/x-socket.io' then it's a Socket simplate.

        If mimetype is 'text/*' or 'application/json' then we look for page
        breaks (^L). If there aren't any page breaks then it's a static file.
        If it has at least one page break then it's a dynamic simplate (either
        Template or JSON).

        For all other mimetypes we sniff the first few bytes of the file. If it
        looks Python-y then it's a Template simplate, otherwise it's a static
        file. What looks Python-y? Triple quotes for a leading docstring, or
        the beginning of an import statement ("from" or "import").
    
    """

    is_dynamic = True

    if mimetype == 'application/x-socket.io':
        
        # *.sock files are always dynamic.

        pass

    elif mimetype.startswith('text/') or mimetype == 'application/json':

        # For text formats we can perform a highly accurate test for
        # dynamicity.

        c = lambda s: s in raw
        is_dynamic = c("") or c("^L")

    else:

        # For binary formats we must rely on a less-accurate test. This is
        # because a binary file can have s in it without being a resource--
        # and I've actually seen, in the wild, a file with exactly two s. So
        # we sniff the first few bytes.

        s = lambda s: raw.startswith(s)
        is_dynamic = s('"""') or s('import') or s('from')

    if not is_dynamic:
        Class = StaticResource
    elif mimetype == 'application/json':
        Class = JSONResource
    elif mimetype == 'application/x-socket.io':
        Class = SocketResource
    else:
        Class = TemplateResource

    return Class


def load(request, modtime):
    """Given a Request and a modtime, return a Resource object (w/o caching).
    """

    # Load bytes.
    # ===========
    # We work with resources exclusively as bytestrings. Any unicode objects
    # passed in by the user as {{ expressions }} in Resources will be encoded
    # with UTF-8 by Tornado.

    raw = open(request.fs, 'rb').read()
    
   
    # Compute a mimetype.
    # ===================

    mimetype = mimetypes.guess_type(request.fs, strict=False)[0]
    if mimetype is None:
        mimetype = request.default_mimetype


    # Compute and instantiate a class.
    # ================================
    # An instantiated resource is compiled as far as we can take it.
 
    Class = get_resource_class(raw, mimetype)
    resource = Class(request.website, request.fs, raw, mimetype, modtime)
    return resource


def get(request):
    """Given a Request, return a Resource object (with caching).

    We need the request because it carries default_mimetype.

    """

    # XXX This is not thread-safe. It used to be, but then I simplified it
    # when I switched to diesel. Now that we have multiple engines, some of
    # which are threaded, we need to make this thread-safe again.
    
    # Get a cache Entry object.
    # =========================

    if request.fs not in __cache__:
        entry = Entry()
        __cache__[request.fs] = entry

    entry = __cache__[request.fs]


    # Process the resource.
    # =====================

    modtime = os.stat(request.fs)[stat.ST_MTIME]
    if entry.modtime == modtime:                            # cache hit
        if entry.exc is not None:
            raise entry.exc
    else:                                                   # cache miss
        try:
            entry.resource = load(request, modtime)
        except:     # capture any Exception
            entry.exc = ( LoadError(traceback.format_exc())
                        , sys.exc_info()[2]
                         )
        else:       # reset any previous Exception
            entry.exc = None 

        entry.modtime = modtime
        if entry.exc is not None:
            raise entry.exc[0] # TODO Why [0] here, and not above?


    # Return
    # ======
    # The caller must take care to avoid mutating any namespace dictionary at 
    # entry.resource.pages[0].

    return entry.resource
